

from ..qt_compat import QtGui, QtCore, QtWidgets

import time

import h5py as h5
import numpy as np
import logging as log

import os
import matplotlib

from .. import actions as ac

from scipy.ndimage.interpolation import zoom
from scipy import ndimage as ndi
from skimage.morphology import disk

from ..plugins import Plugin
from ..plugins.visualization import LayersMenu, ContrastMenu
from ..core import DataModel, LayerManager, LabelManager, Launcher
from .mpl_widgets import PerspectiveCanvas

from .base import TComboBox, PLineEdit, HWidgets, HSize3D, ULabel, OddSlider, \
                  SourceCombo

from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT
from skimage.draw import line, bezier_curve


class NavigationToolbar(NavigationToolbar2QT):

    home_pressed = QtCore.pyqtSignal()
    toggle_grid = QtCore.pyqtSignal()

    toolitems = (
        (None, None, None, None),
        ('Home', 'Reset original view', 'home', 'home'),
        (None, None, None, None),
        ('Draw', 'Draw ground truth seeds', 'pencil', 'gt'),
        #('Region', 'Grow a region', 'region2', 'region'),
        (None, None, None, None),
        ('Pan', 'Pan axes with left mouse, zoom with right', 'enlarge', 'pan'),
        ('Zoom', 'Zoom to rectangle', 'zoom-in', 'zoom'),
        (None, None, None, None),
        ('Grid', 'Turn on/off the grid', 'grid', 'grid'),
        (None, None, None, None),
        ('Save', 'Save the figure', 'floppy-disk', 'save_figure'),
    )

    icondir = os.path.dirname(os.path.realpath(__file__)) + '/../images/PNG/'

    def __init__(self, canvas, parent=None, locLabel=None):
        super(NavigationToolbar, self).__init__(canvas, parent=parent)
        self.parent = parent
        self.locLabel = locLabel
        self.setOrientation(QtCore.Qt.Vertical)

    def gt(self):
        if self._active == 'GT':
            self._active = None
        else:
            self._active = 'GT'
        self._update_buttons_checked()

        if self._idPress is not None:
            self._idPress = self.canvas.mpl_disconnect(self._idPress)
            self.mode = ''

        if self._idRelease is not None:
            self._idRelease = self.canvas.mpl_disconnect(self._idRelease)
            self.mode = ''

    def region(self):
        if self._active == 'REGION':
            self._active = None
        else:
            self._active = 'REGION'
        self._update_buttons_checked()

        if self._idPress is not None:
            self._idPress = self.canvas.mpl_disconnect(self._idPress)
            self.mode = ''

        if self._idRelease is not None:
            self._idRelease = self.canvas.mpl_disconnect(self._idRelease)
            self.mode = ''

    def _icon(self, name):
        icon = QtGui.QIcon(os.path.abspath(os.path.join(self.icondir, name+'.png')))
        return icon

    def _init_toolbar(self):

        con_button = QtWidgets.QToolButton()
        con_action = QtWidgets.QAction(self._icon('contrast'), '', con_button)
        con_button.setDefaultAction(con_action)
        con_menu = ContrastMenu(self)
        con_button.setMenu(con_menu)
        con_button.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        con_button.setStyleSheet('QToolButton {margin-left: 0px; width: 24px;}'
                                 'QToolButton::menu-indicator {'
                                 '  width: 0px; border: none; image: none;'
                                 '}')
        self.addWidget(con_button)

        viz_button = QtWidgets.QToolButton()
        viz_action = QtWidgets.QAction(self._icon('layers'), '', viz_button)
        viz_button.setDefaultAction(viz_action)
        viz_menu = LayersMenu(self)
        viz_button.setMenu(viz_menu)
        viz_button.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        viz_button.setStyleSheet('QToolButton {margin-left: 0px; width: 24px;}'
                                 'QToolButton::menu-indicator {'
                                 '  width: 0px; border: none; image: none;'
                                 '}')
        self.addWidget(viz_button)

        for text, tooltip_text, image_file, callback in self.toolitems:
            if text is None:
                self.addSeparator()
            else:
                icon = self._icon(image_file)
                a = self.addAction(icon, text, getattr(self, callback))
                self._actions[callback] = a
                if callback in ['zoom', 'pan', 'gt', 'region', 'grid']:
                    a.setCheckable(True)
                if tooltip_text is not None:
                    a.setToolTip(tooltip_text)

        self.buttons = {}

        self.adj_window = None

    def _update_buttons_checked(self):
        self._actions['pan'].setChecked(self._active == 'PAN')
        self._actions['zoom'].setChecked(self._active == 'ZOOM')
        self._actions['gt'].setChecked(self._active == 'GT')
        #self._actions['region'].setChecked(self._active == 'REGION')
        self.parent.show_tools(self._active)

    def grid(self):
        self.toggle_grid.emit()

    def layers(self):
        self.vizmenu.exec_()

    def home(self):
        self.home_pressed.emit()


class SliceViewer(Plugin):
    name = 'Slice Viewer'

    def __init__(self, data=None, ptype=Plugin.Widget):
        super(SliceViewer, self).__init__(ptype=ptype)

        self.DM = DataModel.instance()
        self.layered_canvas = LayeredCanvas()

        self.tool_options = QtWidgets.QWidget()
        hbox = QtWidgets.QHBoxLayout()
        self.tool_options.setLayout(hbox)
        hbox.setContentsMargins(0, 0, 0, 0)

        self.locLabel = QtWidgets.QLabel("", self)
        self.locLabel.setAlignment(
            QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.locLabel.setSizePolicy(
            QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                              QtWidgets.QSizePolicy.Ignored))

        self.toolbar = NavigationToolbar(
            self.layered_canvas.canvas, self, self.locLabel)
        self.toolbar.setMaximumWidth(50)
        self.toolbar.toggle_grid.connect(self.layered_canvas.toggle_grid)
        self.toolbar.home_pressed.connect(self.on_home_pressed)

        self.layered_canvas.toolbar = self.toolbar
        self.source_combo = SourceCombo(listen_DM=True, listen_self=True)

        self.layout.addWidget(HWidgets(self.source_combo,
                                       self.tool_options,
                                       self.locLabel,
                                       stretch=[0, 1, 0]))

        self.layout.addWidget(HWidgets(self.toolbar,
                                       self.layered_canvas,
                                       stretch=[0, 1]))

        self.gt_tool = GtTool()
        self.grow_tool = GrowTool()

    def on_home_pressed(self):
        sy, sx = self.DM.region_shape()[1:]
        self.layered_canvas.ax.set_ylim([sy, 0])
        self.layered_canvas.ax.set_xlim([0, sx])
        self.layered_canvas.redraw()

    def increase_slider(self):
        self.layered_canvas.increase_slider()

    def decrease_slider(self):
        self.layered_canvas.decrease_slider()

    def selectSource(self, source):
        self.layered_canvas.combo_source.setCurrentIndex(source)

    def show_tools(self, tool):
        layout = self.tool_options.layout()
        for i in reversed(range(layout.count())):
            layout.itemAt(i).widget().setParent(None)

        if tool == 'GT':
            layout.addWidget(self.gt_tool)
        elif tool == 'REGION':
            layout.addWidget(self.grow_tool)


class LayeredCanvas(PerspectiveCanvas):

    def __init__(self, *args, **kwargs):
        super(LayeredCanvas, self).__init__(*args, **kwargs)
        self.DM.evmin_changed.connect(self.update_vmin)
        self.DM.evmax_changed.connect(self.update_vmax)

        # self.LM.added.connect(self.replot)
        # self.LM.removed.connect(self.replot)
        self.LM.updated.connect(self.replot)
        self.LM.opacity.connect(self.on_layer_opacity)
        # self.LM.toggled.connect(self.on_layer_toggled)

        self.LBLM.labelUpdated.connect(self.replot)

        self.canvas.mpl_connect('button_press_event', self.canvas_onmousepress)
        self.canvas.mpl_connect('button_release_event',
                                self.canvas_onmouserelease)
        self.canvas.mpl_connect('motion_notify_event', self.canvas_onmousemove)
        self.canvas.mpl_connect('scroll_event', self.canvas_onscroll)

        self.slider.valueChanged.connect(self.update_volume)

        self.idx = self.DM.data_shape[0] // 2
        self.DM.current_idx = self.idx

        self.DM.roi_changed.connect(self.update_axes)


        self.layout().setContentsMargins(0, 0, 0, 0)

        self.current = None

        # Draw!
        self.grid = False
        self.replot()

    def __del__(self):
        self.ax.close()

    def update_axes(self):
        sz, sy, sx = self.DM.active_roi
        self.slider.setMinimum(0)
        self.slider.setMaximum(sz.stop - sz.start - 1)
        self.slider.setValue((sz.stop - sz.start) // 2)
        self.ax.set_ylim([sy.stop - sy.start + 1, -1])
        self.ax.set_xlim([-1, sx.stop - sx.start + 1])

    def get_idx(self):
        return self.idx

    def on_layer_opacity(self, name, level, val):
        idx = self.LM.index(str(name), str(level))
        if idx < 0:
            return
        self.ax.images[idx].set_alpha(val)
        self.redraw()

    def on_layer_toggled(self, name, level, bol):
        self.replot()

    def replot(self):
        super(LayeredCanvas, self).replot()
        ylim, xlim = self.ax.get_ylim(), self.ax.get_xlim()
        valid = ylim != (0., 1.) and xlim != (0., 1.)
        for i, layer in enumerate(self.LM.visible_layers()):
            layer.draw(self.ax, self.idx, i)

        if valid:
            h, w = self.DM.data_shape[1:]
            x = [-2, w + 1, w + 1, -2, -2]
            y = [-2, -2, h + 1, h + 1, -2]
            self.ax.plot(x, y, 'r-', linewidth=5)
            self.ax.set_ylim(ylim)
            self.ax.set_xlim(xlim)

        self.ax.grid(self.grid)
        self.redraw()

    def update_vmin(self, val):
        data = self.LM.get('Data', 'Data')
        data.vmin = val
        if not data.binarize:
            self.ax.images[0].set_clim(vmin=val)
            self.redraw()
        else:
            self.LM.update()

    def update_vmax(self, val):
        data = self.LM.get('Data', 'Data')
        data.vmax = val
        if not data.binarize:
            self.ax.images[0].set_clim(vmax=val)
            self.redraw()
        else:
            self.LM.update()

    def increase_slider(self):
        if self.slider.value() < self.slider.maximum():
            self.slider.setValue(self.slider.value() + 1)

    def decrease_slider(self):
        if self.slider.value() > self.slider.minimum():
            self.slider.setValue(self.slider.value() - 1)

    def update_volume(self, idx):
        self.DM.current_idx = idx
        self.idx = idx
        self.text_idx.setText(str(idx))
        for (layer, image) in zip(self.LM.visible_layers(), self.ax.images):
            layer.update(image, self.idx)
        self.redraw()

    def canvas_onscroll(self, ev):
        if ev.button == 'up':
            self.increase_slider()
        else:
            self.decrease_slider()

    def canvas_onmousepress(self, ev):
        if self.DM.gtselected is None or self.toolbar._active != 'GT':
            return
        if ev.inaxes != self.ax:
            return
        self.pressed = True

        y = int(ev.ydata)
        x = int(ev.xdata)
        self.current = [[y], [x]]

        # Extract correct radius for current scale
        brush_width = self.DM.gtradius * 2 + 1
        bbox = self.ax.get_window_extent().transformed(self.fig.dpi_scale_trans.inverted())
        canvas_width = bbox.width * self.fig.dpi
        xlim = self.ax.get_xlim()
        zoom_width = xlim[1] - xlim[0] - 2
        brush_width *=  canvas_width / zoom_width

        self.canvas.setMouseWidth(brush_width)
        self.canvas.updateMouseLines(ev.y, ev.x)

    def canvas_onmousemove(self, ev):
        if not self.pressed or ev.inaxes != self.ax:
            return
        y = int(ev.ydata)
        x = int(ev.xdata)
        self.current[0] += [y]
        self.current[1] += [x]
        self.canvas.updateMouseLines(ev.y, ev.x)

    def canvas_onmouserelease(self, ev):
        if self.DM.gtselected is None:
            return

        if self.pressed and self.toolbar._active == 'GT':
            self.pressed = False
            self.update_gt(ev)
        elif self.toolbar._active == 'REGION' and ev.inaxes == self.ax:
            self.calculate_region(ev)

        self.current = None
        self.canvas.setMouseLines(None)

    def calculate_region(self, ev):
        y = int(ev.ydata)
        x = int(ev.xdata)
        pos = (self.idx, y, x)
        source, lamda, bbox, region = self.DM.growing_bbox
        if self.DM.attrs('data/{}'.format(source), 'active') != 1:
            QtWidgets.QMessageBox.critical(self, 'Error',
                                       '{} dataset not valid'.format(source))
            return

        if region == 0:
            ldata = self.DM.gtselected['data']
            label = self.DM.gtselected['label']
            ds = self.DM.load_ds('data/{}'.format(source))
            self.launcher.run(ac.grow_voxels, dataset=ds, pos=pos, bbox=bbox,
                              lamda=lamda, annotations=ldata, label=label,
                              caption='Growing region..', cb=self.on_region_grow)
        else:
            QtWidgets.QMessageBox.critical(self, "Error", "Not supported yet.")

    def on_region_grow(self, params):
        indexes, values = params
        self.DM.gtselected['modified'] = True
        level = self.DM.gtselected['level']
        label = self.DM.gtselected['label']
        data = self.DM.gtselected['data']
        self.DM.last_changes = (level, label, indexes, values)
        self.LM.update()

    def update_gt(self, ev):
        data_shape = self.DM.data_shape
        level = self.DM.gtselected['level']
        label = self.DM.gtselected['label']
        plevel = self.DM.gtselected['parent_level']
        plabel = self.DM.gtselected['parent_label']

        t0 = time.time()

        # Get positions
        log.debug('* Smoothing indexes')
        cr, cc = self.current
        pr, pc = [], []
        if self.DM.gtinterpolation == 'bezier':
            for i in range(1, len(cr) - 1, 2):
                line_r, line_c = bezier_curve(cr[i - 1], cc[i - 1],
                                              cr[i], cc[i],
                                              cr[i + 1], cc[i + 1],
                                              2)
                pr.extend(line_r)
                pc.extend(line_c)
            if len(cr) % 2 == 0:
                line_r, line_c = line(cr[-2], cc[-2], cr[-1], cc[-1])
                pr.extend(line_r)
                pc.extend(line_c)
        else:
            for i in range(len(cr) - 1):
                line_r, line_c = line(cr[i], cc[i], cr[i + 1], cc[i + 1])
                pr.extend(line_r)
                pc.extend(line_c)

        pos = np.c_[pr, pc]

        if pos.shape[0] == 0:
            pos = np.c_[cr, cc]

        print("Time Smooth:", time.time() - t0)
        t0 = time.time()

        _, region_maxy, region_maxx = self.DM.region_shape()
        pos[:, 0] = np.clip(pos[:, 0], 0, region_maxy - 1)
        pos[:, 1] = np.clip(pos[:, 1], 0, region_maxx - 1)

        radius = self.DM.gtradius

        ymin, xmin = pos.min(0) - radius
        ymax, xmax = pos.max(0) + radius

        ymin = max(ymin, 0)
        xmin = max(xmin, 0)
        ymax = min(ymax, region_maxy - 1)
        xmax = min(xmax, region_maxx - 1)

        mask2d = np.zeros((ymax-ymin+1, xmax-xmin+1), np.bool)
        mask2d[pos[:, 0] - ymin, pos[:, 1] - xmin] = True
        mask2d = ndi.binary_dilation(mask2d, disk(radius).astype(np.bool))

        slice_z = slice(self.idx, self.idx+1)
        slice_y = slice(ymin, ymax+1)
        slice_x = slice(xmin, xmax+1)

        pos = np.column_stack(np.where(mask2d))

        print("Time Dilate:", time.time() - t0)

        t0 = time.time()

        log.debug('* Calculating indexes')
        if self.DM.gtlevel == 1:

            t0 = time.time()

            sv = self.DM.load_slices(self.DM.svlabels, slice_z, slice_y, slice_x,
                                     apply_roi=True)
            sv = sv[0, pos[:, 0], pos[:, 1]]
            sv = np.unique(sv)
            if np.isnan(sv).any() or (sv < 0).any():
                self.launcher.show_error('SuperVoxels are not created for this ROI')
                return

            total_idx = np.array([], np.int32)

            with h5.File(self.DM.ds_path(self.DM.svtable), 'r') as t1, \
                 h5.File(self.DM.ds_path(self.DM.svindex), 'r') as t2:

                svtable = t1['data']
                svindex = t2['data']

                for v in sv:
                    if v == self.DM.svtotal - 1:
                        slc = slice(svtable[v], None)
                    else:
                        slc = slice(svtable[v], svtable[v+1])
                    total_idx = np.append(total_idx, svindex[slc])

            indexes = np.column_stack(np.unravel_index(total_idx, self.DM.data_shape))
            if self.DM.data_shape != self.DM.region_shape():
                log.debug('* Filtering indexes')
                mask = np.ones(indexes.shape[0], np.bool)
                for i, axis in enumerate(self.DM.active_roi):
                    curr = indexes[:, i]
                    mask[curr < axis.start] = False
                    mask[curr >= axis.stop] = False
                indexes = indexes[mask]

            zmin, ymin, xmin = indexes.min(0)
            zmax, ymax, xmax = indexes.max(0)
            indexes[:, 0] -= zmin
            indexes[:, 1] -= ymin
            indexes[:, 2] -= xmin
            apply_roi = False
        elif self.DM.gtlevel == 2:
            mv = self.DM.load_slices(self.DM.mvlabels, slice_z, slice_y, slice_x)
            mv = mv[0, pos[:, 0], pos[:, 1]]
            mv = np.unique(mv)
            if np.isnan(mv).any() or (mv < 0).any():
                self.launcher.show_error('MegaVoxels are not created for this ROI')
                return

            total_idx = np.array([], np.int32)

            with h5.File(self.DM.ds_path(self.DM.mvtable), 'r') as t1, \
                 h5.File(self.DM.ds_path(self.DM.mvindex), 'r') as t2:

                mvtable = t1['data']
                mvindex = t2['data']

                for v in mv:
                    if v == self.DM.mvtotal - 1:
                        slc = slice(mvtable[v], None)
                    else:
                        slc = slice(mvtable[v], mvtable[v+1])
                    total_idx = np.append(total_idx, mvindex[slc])

            indexes = np.column_stack(np.unravel_index(total_idx, self.DM.data_shape))
            if self.DM.data_shape != self.DM.region_shape():
                log.debug('* Filtering indexes')
                mask = np.ones(indexes.shape[0], np.bool)
                for i, axis in enumerate(self.DM.active_roi):
                    curr = indexes[:, i]
                    mask[curr < axis.start] = False
                    mask[curr >= axis.stop] = False
                indexes = indexes[mask]

            zmin, ymin, xmin = indexes.min(0)
            zmax, ymax, xmax = indexes.max(0)
            indexes[:, 0] -= zmin
            indexes[:, 1] -= ymin
            indexes[:, 2] -= xmin
            apply_roi = False
        else:
            zmin = zmax = self.idx
            indexes = np.empty((pos.shape[0], 3), np.int32)
            indexes[:, 0] = 0
            indexes[:, 1:] = pos
            apply_roi = True

        print("Time Get Index:", time.time() - t0)

        t0 = time.time()

        slice_z = slice(zmin, zmax+1)
        slice_y = slice(ymin, ymax+1)
        slice_x = slice(xmin, xmax+1)

        if plevel is not None and plevel >= 0 and plabel >= 0:
            plevel = self.LBLM.dataset(plevel)
            pdata = self.DM.load_slices(plevel, slice_z, slice_y, slice_x, apply_roi=apply_roi)
            mask = (pdata == plabel)
            valid = mask[indexes[:, 0], indexes[:, 1], indexes[:, 2]]
            indexes = indexes[valid]

        hdata = self.DM.load_slices(level, slice_z, slice_y, slice_x, apply_roi=apply_roi)
        values = hdata[indexes[:, 0], indexes[:, 1], indexes[:, 2]]
        hdata[indexes[:, 0], indexes[:, 1], indexes[:, 2]] = label
        self.DM.write_slices(level, hdata, slice_z, slice_y, slice_x, apply_roi=apply_roi)
        self.DM.last_changes = (level, (slice_z, slice_y, slice_x),
                                indexes, values, apply_roi)

        print("Time Replace Data:", time.time() - t0)

        log.debug('* Done')
        self.update_volume(self.idx)

    def toggle_grid(self):
        self.grid = not self.grid
        self.ax.grid(self.grid)
        self.redraw()


class GtTool(HWidgets):

    def __init__(self):
        items = ['Voxels', 'SuperVoxels', 'MegaVoxels']
        self.DM = DataModel.instance()
        self.launcher = Launcher.instance()
        self.alevel_combo = TComboBox('Annotation Level:', items,
                                      selected=self.DM.gtlevel)
        self.radius_slider = OddSlider('', vmin=1, vmax=31, current=1,
                                       interval=2, lblwidth=30)
        self.interp_type = TComboBox('Interpolation:', ['linear', 'bezier'])
        super(GtTool, self).__init__(self.alevel_combo, 'Width:',
                                     self.radius_slider, self.interp_type, None,
                                     stretch=[0, 0, 0, 0, 1])

        self.alevel_combo.currentIndexChanged.connect(self.on_annotation_level)
        self.radius_slider.valueChanged.connect(self.on_radius)
        self.interp_type.currentIndexChanged.connect(self.on_interpolation)

    def on_annotation_level(self, idx):
        if idx == 1 and self.DM.svlabels is None:
            self.launcher.error.emit('SuperVoxels need to be created first')
            self.alevel_combo.setCurrentIndex(self.DM.gtlevel)
        elif idx == 2 and self.DM.mvlabels is None:
            self.launcher.error.emit('MegaVoxels need to be created first')
            self.alevel_combo.setCurrentIndex(self.DM.gtlevel)
        else:
            self.DM.gtlevel = idx

    def on_interpolation(self):
        self.DM.gtinterpolation = self.interp_type.currentText()

    def on_radius(self, idx):
        self.DM.gtradius = int(self.radius_slider.value()) // 2


class GrowTool(HWidgets):

    def __init__(self):
        self.DM = DataModel.instance()
        self.launcher = Launcher.instance()

        z, y, x = self.DM.growing_bbox[2]
        self.bboxz = PLineEdit(z, parse=int)
        self.bboxz.setMaximumWidth(50)
        self.bboxy = PLineEdit(y, parse=int)
        self.bboxy.setMaximumWidth(50)
        self.bboxx = PLineEdit(x, parse=int)
        self.bboxx.setMaximumWidth(50)

        self.bbox_lamda = PLineEdit(self.DM.growing_bbox[1], parse=float)
        self.bbox_lamda.setMaximumWidth(50)
        items = ['data', 'stretch', 'smooth', 'denoised']
        self.bbox_source = TComboBox('Source:', items,
                                     selected=items.index(self.DM.growing_bbox[0]))
        items = ['Voxels', 'SuperVoxels', 'MegaVoxels']
        self.alevel_combo = TComboBox('Annotation Level:', items,
                                      selected=self.DM.growing_bbox[3])
        super(GrowTool, self).__init__(self.alevel_combo, 'ROI:',
                                       ULabel('z'), self.bboxz,
                                       ULabel('y'), self.bboxy,
                                       ULabel('x'), self.bboxx,
                                       'Lambda:', self.bbox_lamda,
                                       self.bbox_source, None,
                                       stretch=[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1])

        self.alevel_combo.currentIndexChanged.connect(self.on_change)
        self.bboxz.textEdited.connect(self.on_change)
        self.bboxy.textEdited.connect(self.on_change)
        self.bboxx.textEdited.connect(self.on_change)
        self.bbox_lamda.textChanged.connect(self.on_change)
        self.bbox_source.currentIndexChanged.connect(self.on_change)

    def on_change(self):
        bbox = [self.bboxz.value(), self.bboxy.value(), self.bboxx.value()]
        lamda = self.bbox_lamda.value()
        source = self.bbox_source.currentText()
        level = self.alevel_combo.currentIndex()

        if level == 1 and self.DM.svlabels is None:
            self.launcher.error.emit('SuperVoxels need to be created first')
            self.alevel_combo.setCurrentIndex(0)
            return
        elif level == 2 and self.DM.mvlabels is None:
            self.launcher.error.emit('MegaVoxels need to be created first')
            self.alevel_combo.setCurrentIndex(0)
            return

        if not self.DM.has_ds('data/{}'.format(source)):
            self.bbox_source.setCurrentIndex(0)
            errmsg = 'Dataset {} has to be created first'.format(source)
            QtWidgets.QMessageBox.critical(self, 'Error', errmsg)
            return

        self.DM.growing_bbox = [source, lamda, bbox, level]




import math
import numpy as np

from ..qt_compat import QtGui, QtCore, QtWidgets

import six
import logging as log
import time
import re
import shutil

from .slice_viewer import SliceViewer
from .label_partitioning import LabelSplitter
from .level_statistics import LevelStats
from .preloader import PreWidget
from .data_loader import LoadWidget
from .compare_levels import QuantitativeAnalysis

from ..core import Launcher
from ..core import DataModel, LayerManager, LabelManager
from ..plugins import Plugin
from ..plugins import ROI, SuperRegions, Annotations, Visualization,\
                      Training, FeatureChannels, Export
from ..widgets import RCheckBox, HeaderButton, HWidgets, BLabel, PicButton, \
                      ComboDialog

from time import sleep

from .. import actions as ac

import os
import fnmatch
import h5py as h5

class QPlainTextEditLogger(log.Handler):
    def __init__(self, parent):
        super(QPlainTextEditLogger, self).__init__()
        self.widget = parent

    def emit(self, record):
        self.widget.setText(record.message)

    def write(self, m):
        pass

class MainWindow(QtWidgets.QMainWindow):

    def __init__(self, title="SuRVoS: Super-Region Volume Segmentation workbench"):
        super(MainWindow, self).__init__()

        self.DM = None

        self.pre_widget = PreWidget()
        self.load_widget = LoadWidget()
        self.main_widget = MainWidget()
        self.main_container = QtWidgets.QStackedWidget()
        self.main_container.addWidget(self.pre_widget)
        self.main_container.addWidget(self.load_widget)
        self.main_container.addWidget(self.main_widget)
        self.setCentralWidget(self.main_container)

        self.setWindowTitle(title)

        self.overlay = Overlay(parent=self)
        self.launcher = Launcher.instance()
        self.launcher.pre.connect(self.overlay.pre)
        self.launcher.post.connect(self.overlay.post)
        self.launcher.error.connect(self.on_error)

        self.status = self.statusBar()
        self.edit = QtWidgets.QLabel()
        self.edit.setStyleSheet('color: #fefefe; padding-left: 10px;');
        self.edit.setFixedHeight(25)
        self.status.insertPermanentWidget(0, self.edit, 1)

        self.loghandler = QPlainTextEditLogger(self.edit)
        log.getLogger().addHandler(self.loghandler)

        menubar = QtWidgets.QMenuBar()
        menubar.setStyleSheet('QMenuBar {padding: 6px;}')
        menubar.setContentsMargins(20,10,0,0)
        self.setMenuBar(menubar)
        fileMenu = menubar.addMenu('&File')
        helpMenu = menubar.addMenu('&Help')

        self.openAction = QtWidgets.QAction('&Open Dataset..', self)
        self.openAction.setShortcut('Ctrl+O')
        self.openAction.setStatusTip('Open Dataset (.rec, .hdf5, .tiff)')
        self.openAction.triggered.connect(self.load_data_view)
        self.pre_widget.open.clicked.connect(self.load_data_view)

        self.loadAction = QtWidgets.QAction('&Load Workspace..', self)
        self.loadAction.setShortcut('Ctrl+L')
        self.loadAction.setStatusTip('Load previously created workspace.')
        self.loadAction.triggered.connect(self.load_workspace)
        self.pre_widget.load.clicked.connect(self.load_workspace)

        self.saveAction = QtWidgets.QAction('&Save annotations', self)
        self.saveAction.setShortcut('Ctrl+S')
        self.saveAction.setStatusTip('Save and backup current annotations.')
        self.saveAction.triggered.connect(self.main_widget.save_backup)
        self.saveAction.setEnabled(False)

        exitAction = QtWidgets.QAction('&Exit', self)
        exitAction.setShortcut('Ctrl+Q')
        exitAction.setStatusTip('Exit application')
        exitAction.triggered.connect(QtWidgets.qApp.quit)

        fileMenu.addAction(self.openAction)
        fileMenu.addAction(self.loadAction)
        fileMenu.addAction(self.saveAction)
        fileMenu.addSeparator()
        fileMenu.addAction(exitAction)

        # HEEEEEELP
        self.docAction = QtWidgets.QAction('Documentation..', self)
        self.docAction.triggered.connect(self.open_docs)
        self.bugAction = QtWidgets.QAction('Issues and Bugs..', self)
        self.bugAction.triggered.connect(self.open_issues)
        self.aboutAction = QtWidgets.QAction('About', self)

        helpMenu.addAction(self.docAction)
        helpMenu.addAction(self.bugAction)
        helpMenu.addAction(self.aboutAction)

        # LOADER
        self.load_widget.data_loaded.connect(self.load_data)
        self.load_widget.cancel.clicked.connect(self.show_main)

        self.DM = DataModel.instance()
        self.DM.main_window = self
        self.LM = LayerManager.instance()
        self.LBLM = LabelManager.instance()

    def show_main(self):
        self.main_container.setCurrentIndex(0)

    def setGPU(self, gpu):
        self.DM.selected_gpu = gpu

    def open_docs(self):
        docURL = 'https://github.com/DiamondLightSource/SuRVoS/wiki'
        QtGui.QDesktopServices.openUrl(QtCore.QUrl(docURL))

    def open_issues(self):
        bugURL = 'https://github.com/DiamondLightSource/SuRVoS/issues'
        QtGui.QDesktopServices.openUrl(QtCore.QUrl(bugURL))

    def load_data_view(self):
        self.main_container.setCurrentIndex(1)
        self.load_widget.winput.open_dialog()        

    def load_data(self, wdir):
        self.main_widget.prepare_data(wdir)
        self.main_container.setCurrentIndex(2)

    def load_workspace(self, path=None):
        if type(path) not in six.string_types:
            msg = "Select the workspace folder"
            flags = QtWidgets.QFileDialog.ShowDirsOnly
            path = QtWidgets.QFileDialog.getExistingDirectory(self, msg, '.', flags)
            if path is None or len(path) == 0:
                return

        if self.main_widget.load_workspace(path):
            self.main_container.setCurrentIndex(2)

    def on_error(self, msg):
        QtWidgets.QMessageBox.critical(self, "Error", msg)
        self.overlay.post()

    def addWidget(self, widget, ptype=None):
        self.centralWidget().addWidget(widget, ptype)

    def resizeEvent(self, event):
        self.overlay.resize(event.size())
        event.accept()

    def closeEvent(self, event):
        log.info("\n### Destroying ###")
        event.accept()

    def keyPressEvent(self, event):
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if event.key() == QtCore.Qt.Key_Escape:
            self.setFocus()
        elif event.key() >= QtCore.Qt.Key_1 and \
                event.key() < QtCore.Qt.Key_1 + self.main_widget.leftPanel.count():
            index = event.key() - QtCore.Qt.Key_1
            if self.main_widget.leftPanel.isTabEnabled(index):
                self.main_widget.leftPanel.setCurrentIndex(index)
        elif modifiers == QtCore.Qt.ControlModifier and \
                event.key() == QtCore.Qt.Key_Z:
            # CTRL + Z
            log.debug('* DEBUG: CTRL+Z clicked')
            if self.DM.last_changes is not None:
                ds, slices, indexes, values, active_roi = self.DM.last_changes
                slice_z, slice_y, slice_x = slices

                hdata = self.DM.load_slices(ds, slice_z, slice_y, slice_x,
                                            apply_roi=active_roi)
                prev_values = hdata[indexes[:, 0], indexes[:, 1], indexes[:, 2]]
                hdata[indexes[:, 0], indexes[:, 1], indexes[:, 2]] = values
                self.DM.write_slices(ds, hdata, slice_z, slice_y, slice_x,
                                     apply_roi=active_roi)
                self.DM.last_changes = (ds, (slice_z, slice_y, slice_x),
                                        indexes, prev_values, active_roi)
                self.LM.update()
                log.info('+ Done')

    def mousePressEvent(self, event):
        self.setFocus()

class MainWidget(QtWidgets.QWidget):

    completed = QtCore.pyqtSignal()

    def __init__(self, title='QMainWindow'):
        super(MainWidget, self).__init__()
        self.setWindowTitle(title)

        self.pluginCount = 0

        self.setLayout(QtWidgets.QHBoxLayout())
        self.splitter = QtWidgets.QSplitter()
        self.layout().setContentsMargins(20,20,20,20)
        self.layout().addWidget(self.splitter)

        dirp = os.path.dirname(os.path.realpath(__file__))
        self.leftPanel = QtWidgets.QTabWidget(self)
        self.leftPanel.setTabPosition(QtWidgets.QTabWidget.West)
        self.leftPanel.setMaximumWidth(400)

        self.centerPanel = QtWidgets.QTabWidget()
        self.centerPanel.setTabPosition(QtWidgets.QTabWidget.West)

        self.splitter.addWidget(self.leftPanel)
        self.splitter.addWidget(self.centerPanel)

        self.launcher = Launcher.instance()
        self.DM = DataModel.instance()
        self.LM = LayerManager.instance()
        self.LBLM = LabelManager.instance()

        self.slice_viewer = None

    def __add__(self, plugin):
        return self.add_widget(plugin)

    def prepare_data(self, wdir):
        log.info("Workspace directory: {}".format(wdir))
        self.on_workspace(wdir)
        self.launcher.run(ac.load_data, data=self.DM.tmp_data, stats=self.DM.tmp_stats,
                          cb=self.on_data_loaded, caption='Loading data..')

    def on_data_loaded(self, result):
        del self.DM.tmp_data; self.DM.tmp_data = None
        self.DM.data = '/data'
        self.DM.data_shape = self.DM.ds_shape('/data')
        self.DM.active_roi = (slice(0, self.DM.data_shape[0]),
                              slice(0, self.DM.data_shape[1]),
                              slice(0, self.DM.data_shape[2]))
        attrs = self.DM.attrs('/data')

        # Left column
        self.visualization = Visualization()
        self.roi = ROI()
        self.SuperRegions = SuperRegions()
        self.voxel_channels = FeatureChannels()
        self.annotations = Annotations()
        self.training = Training()
        self.export = Export()

        self.addWidget(self.visualization, pvisible=True)

        self.addWidget(self.roi, pvisible=True, enabled=True, after=self)
        self.addWidget(self.voxel_channels, pvisible=False, \
                       enabled=True, after=self)
        self.addWidget(self.SuperRegions, pvisible=False, \
                       enabled=True, after=self)
        self.addWidget(self.annotations, pvisible=False, \
                       enabled=True, after=self)
        self.addWidget(self.training, pvisible=False, \
                       enabled=True, after=self)
        self.addWidget(self.export, pvisible=False, \
                       enabled=True, after=self)

        # Midle
        self.slice_viewer = SliceViewer()
        self.addWidget(self.slice_viewer)
        self.label_splitter = LabelSplitter()
        self.addWidget(self.label_splitter)
        self.level_statistics = LevelStats()
        self.addWidget(self.level_statistics)
        self.compare_segmentations = QuantitativeAnalysis()
        self.addWidget(self.compare_segmentations)

        self.LM.clear()
        self.LM.addLayer(self.DM.data, 'Data', level='Data', cmap=attrs['cmap'],
                         vmin=attrs['evmin'], vmax=attrs['evmax'],
                         orient=self.DM.Axial)
        self.LM.update()

    def on_workspace(self, wdir):
        self.DM.wspath = wdir

        self.DM.main_window.openAction.setEnabled(False)
        self.DM.main_window.loadAction.setEnabled(False)
        self.DM.main_window.saveAction.setEnabled(True)

    def load_workspace(self, ws_path):
        self.DM.wspath = ws_path
        if self.on_load_workspace():
            self.on_workspace(ws_path)
            return True
        else:
            errmsg = 'No valid workspace found'
            QtWidgets.QMessageBox.critical(self, 'Error', errmsg)
        return False

    def save_backup(self):
        self.launcher.setup('+ Saving backups for levels: {}'
                            .format(self.LBLM.levels()))
        for level in self.LBLM.levels():
            dataset = self.LBLM.dataset(level)
            log.info('+ Backing up [Level {}]: {}'.format(level, dataset))

            out_dataset = '{}.bak'.format(self.DM.ds_path(dataset))

            if os.path.isfile(out_dataset):
                err_msg = 'Backup for [Level {}] already exists, ' \
                          'do you want to overwrite it?'.format(level)
                ans = QtWidgets.QMessageBox.question(self, "Error", err_msg,
                                                 QtWidgets.QMessageBox.Yes,
                                                 QtWidgets.QMessageBox.No)
                if ans == QtWidgets.QMessageBox.No:
                    log.info('  * Skipping.')
                    continue

            data = self.DM.load_ds(dataset)
            attrs = self.DM.attrs(dataset)

            with h5.File(out_dataset, 'w') as f:
                out_ds = f.create_dataset('data', data=data)
                for k, v in attrs.items():
                    out_ds.attrs[k] = v

            log.info('  * Done.')
        self.launcher.cleanup()

    def on_load_workspace(self):
        pdata = []

        if self.DM.has_ds('/data'):
            attrs = self.DM.attrs('/data')
            self.on_data_loaded(True)
            self.roi.loadROIs()
        else:
            return False

        if self.DM.has_grp('channels'):
            log.info('+ Loading Feature Channels')
            for fidx, desc in sorted(self.DM.available_channels(filter_active=False)):
                attrs = self.DM.attrs(desc)
                keys = list(attrs.keys())
                active = attrs['active']
                fname = attrs['feature_name']
                ftype = attrs['feature_type']
                keys.remove('feature_idx')
                keys.remove('feature_name')
                keys.remove('feature_type')
                keys.remove('active')
                params = {k: attrs[k] for k in keys}
                name = desc.split('/')[1]
                log.info('* Channel {} {}'.format(fname, ftype))
                self.voxel_channels.load_channel(fidx, name, ftype, active, params)

        if self.DM.has_ds('supervoxels/supervoxels') and \
                self.DM.has_ds('supervoxels/supervoxels_idx') and \
                self.DM.has_ds('supervoxels/supervoxels_table') and \
                self.DM.has_ds('supervoxels/graph_edges') and \
                self.DM.has_ds('supervoxels/graph_edge_weights'):
            log.info('+ Loading SuperVoxels')
            svds = 'supervoxels/supervoxels'
            svidx = 'supervoxels/supervoxels_idx'
            svtable = 'supervoxels/supervoxels_table'
            attrs = self.DM.attrs(svds)
            source = attrs['source']
            sp_shape = attrs['sp_shape']
            spacing = attrs['spacing']
            compactness = attrs['compactness']
            sptotal = attrs['num_supervoxels']
            self.SuperRegions.load_supervoxels(svds, svidx, svtable,
                                                source, sp_shape, spacing,
                                                compactness, sptotal)

        if self.DM.has_ds('megavoxels/megavoxels') and \
                self.DM.has_ds('megavoxels/megavoxels_idx') and \
                self.DM.has_ds('megavoxels/megavoxels_table'):
            log.info('+ Loading MegaVoxels')
            mvds = 'megavoxels/megavoxels'
            mvidx = 'megavoxels/megavoxels_idx'
            mvtable = 'megavoxels/megavoxels_table'
            attrs =  self.DM.attrs(mvds)
            source = attrs['source']
            lamda = attrs['lamda']
            nbins = attrs['nbins']
            gamma = attrs['gamma']
            mptotal = attrs['num_megavoxels']
            self.SuperRegions.load_megavoxels(mvds, mvidx, mvtable,
                                               source, lamda, nbins, gamma,
                                               mptotal)

        if self.DM.has_grp('annotations'):
            for levelid, dataset, active in sorted(self.DM.available_annotations()):
                if active != True:
                    log.info('+ Found [Level {}] -- skipping'.format(levelid))
                    self.LBLM.foundLevel(levelid)
                    continue
                self.LBLM.loadLevel(levelid, dataset)

        log.info('+ Ready.')
        self.LM.update()
        return True


    def addWidget(self, plugin, ptype=None, pvisible=True, enabled=True, after=None):
        ptype = ptype or plugin.ptype

        if ptype == Plugin.Widget:
            self.centerPanel.addTab(plugin, plugin.name)
            self.centerPanel.currentChanged.connect(plugin.on_tab_changed)
        elif ptype == Plugin.Plugin:
            plugin.tab_idx = self.leftPanel.count()
            self.leftPanel.currentChanged.connect(plugin.on_tab_changed)
            self.leftPanel.addTab(plugin, plugin.name)
            self.pluginCount += 1
        return self


class Overlay(QtWidgets.QWidget):

    def __init__(self, title='Loading...', parent=None):
        QtWidgets.QWidget.__init__(self, parent)
        palette = QtGui.QPalette(self.palette())
        palette.setColor(palette.Background, QtCore.Qt.transparent)

        self.setPalette(palette)
        self.title = title
        self.hide()
        self.timer = None

    def pre(self, caption):
        self.title = caption
        self.show()
        time.sleep(0.5)

    def paintEvent(self, event):
        painter = QtGui.QPainter()
        painter.begin(self)
        painter.setRenderHint(QtGui.QPainter.Antialiasing)
        painter.fillRect(event.rect(), QtGui.QBrush(QtGui.QColor(255, 255, 255, 127)))
        painter.setPen(QtGui.QPen(QtCore.Qt.NoPen))

        for i in range(6):
            if self.counter == i:
                painter.setBrush(QtGui.QBrush(QtGui.QColor(0, 0, 50)))
            else:
                painter.setBrush(QtGui.QBrush(QtGui.QColor(127, 127, 127)))
            painter.drawEllipse(
                self.width()/2 + 30 * math.cos(2 * math.pi * i / 6.0) - 10,
                self.height()/2 + 30 * math.sin(2 * math.pi * i / 6.0) - 10,
                20, 20)

            font = painter.font()
            font.setPointSize(16)
            painter.setFont(font)
            painter.setPen(QtGui.QColor(80, 80, 80));
            painter.drawText(self.width()//2-200, self.height()//2-40, 400, 200,\
                             QtCore.Qt.AlignCenter, self.title)

        painter.end()

    def showEvent(self, event):
        self.counter = 0
        self.timer = self.startTimer(80)

    def timerEvent(self, event):
        self.counter += 1
        self.counter %= 6
        self.update()

    def post(self):
        log.info('+ Ready')
        self.hide()

    def resizeEvent(self, ev):
        #self.stop.move(self.width()//2-50, self.height()//2 + 100)
        pass

    def cancel_task(self):
        Launcher.instance().terminate()




from ..qt_compat import QtGui, QtCore, QtWidgets


class Plugin(QtWidgets.QScrollArea):

    name = "Plugin"
    header = None

    Widget = 'widget'
    Plugin = 'plugin'

    completed = QtCore.pyqtSignal(bool)

    def __init__(self, ptype=Plugin, height=0, width=300, dock='left', parent=None):

        super(Plugin, self).__init__(parent=parent)

        self.ptype = ptype
        self.container = QtWidgets.QWidget()
        self.container.setWindowTitle(self.name)
        self.setWidget(self.container)
        self.setWidgetResizable(True)

        self.layout = QtWidgets.QGridLayout(self.container)
        self.layout.setAlignment(QtCore.Qt.AlignTop)
        if ptype == self.Plugin:
            self.layout.setContentsMargins(0,10,0,0)
        self.container.setLayout(self.layout)
        self.row = 0
        self.tab_idx = 0

    def setup(self):
        self.header.setEnabled(True)
        self.header.setVisible(True)

    def addWidget(self, widget, align=QtCore.Qt.AlignTop):
        self.layout.addWidget(widget, self.row, 0, align)
        self.row += 1

    def __add__(self, widget):
        self.addWidget(widget)
        return self

    def toggle(self):
        self.setVisible(not self.isVisible())

    def value(self):
        pass

    def on_focus(self):
        pass

    def on_tab_changed(self, idx):
        pass





import numpy as np
import networkx as nx
import logging as log

from ..lib.convolutions import make_3d_gaussian

from scipy.ndimage.measurements import label as splabel
from scipy.ndimage.morphology import binary_closing, binary_dilation, \
                                     binary_erosion, binary_opening, \
                                     binary_fill_holes
from skimage.morphology import diamond, octahedron
from skimage.draw import circle

from ..core import DataModel


DM = DataModel.instance()

def refine_label(data=None, label=None, method=None, radius=1, slide=None):
    ds = data
    data = DM.load_slices(ds)
    rshape = DM.region_shape()

    zmin = DM.active_roi[0].start
    ymin = DM.active_roi[1].start
    xmin = DM.active_roi[2].start

    if method == 'fill_holes':
        radius = 1

    if type(slide) == str:
        mask = (data == label)
        if slide == '3D':
            msize = max(rshape)
            selem = octahedron(radius)
        else:
            msize = max(rshape[1:])
            selem = diamond(radius)
    else:
        msize = max(rshape[1:])
        mask = (data[slide] == label)
        selem = diamond(radius)

    if radius > np.sqrt(msize):
        raise Exception('Radius too large')

    funcs = {
        'dilation' : binary_dilation,
        'erosion' : binary_erosion,
        'opening' : binary_opening,
        'closing' : binary_closing,
        'fill_holes' : binary_fill_holes
    }

    if method not in funcs:
        log.info('+ Refinement {} not supported'.format(method))
        return None

    log.info('+ Performing {} refinement ({})'.format(method, slide))
    if slide != '2D' and mask.any():
        result = funcs[method](mask, structure=selem)
    else:
        result = np.zeros(data.shape, dtype=np.bool)
        f = funcs[method]
        for i in range(data.shape[0]):
            if mask[i].any():
                result[i] = f(mask[i], structure=selem)

    log.info('+ Calculating changes..')
    changes = (result != mask)
    if type(slide) == str:
        values = data[changes]
        data[mask] = -1
        data[result] = label
        changes = np.column_stack(np.where(changes)) + np.array([zmin, ymin, xmin], np.int32)
    else:
        values = data[slide, changes]
        data[slide, mask] = -1
        data[slide, result] = label
        changes = np.column_stack(np.where(changes[None, ...])) + np.array([zmin+slide, ymin, xmin], np.int32)

    DM.write_slices(ds, data)

    log.info('+ done.')
    return changes, values



def save_threshold(source=None, vmin=None, vmax=None, level=None, label=None):
    data = DM.load_slices(source)
    target = DM.load_slices(level)
    mask = (data > vmin) & (data < vmax)
    target[mask] = label
    DM.write_slices(level, target)



import h5py as h5
import numpy as np
import logging as log

import networkx as nx
from sklearn.decomposition import PCA, RandomizedPCA, IncrementalPCA
from sklearn.preprocessing import StandardScaler
from sklearn.random_projection import SparseRandomProjection
from sklearn.ensemble import ExtraTreesClassifier, \
                             RandomForestClassifier, \
                             AdaBoostClassifier,\
                             GradientBoostingClassifier
from sklearn.svm import SVC
from sklearn.cluster import MiniBatchKMeans

from sklearn.linear_model import SGDClassifier
from scipy.stats import entropy

from skimage.segmentation import relabel_sequential

from ..lib._spencoding import _sp_labels
from ..lib.spencoding import spmeans, sphist, spstats
from ..lib.spgraph import aggregate_neighbors
from ..lib._qpbo import solve_binary, solve_aexpansion

from ..core import DataModel


DM = DataModel.instance()


def project_features(X, mode='random_projection'):
    if mode == 'random_projection':
        transformer = SparseRandomProjection(n_components=X.shape[1])
    elif mode == 'standard':
        transformer = StandardScaler()
    elif mode == 'pca':
        transformer = PCA(n_components='mle', whiten=True)
    elif mode == 'random_pca':
        transformer = RandomizedPCA(whiten=True)
    else:
        raise Error('Projection {} not available'.format(mode))

    return transformer.fit(X)


def obtain_classifier(clf_p):
    if clf_p['clf'] == 'ensemble':
        mode = 'ensemble'
        if clf_p['type'] == 'rf':
            clf = RandomForestClassifier(n_estimators=clf_p['n_estimators'],
                                         max_depth=clf_p['max_depth'],
                                         n_jobs=clf_p['n_jobs'])
        elif clf_p['type'] == 'erf':
            clf = ExtraTreesClassifier(n_estimators=clf_p['n_estimators'],
                                       max_depth=clf_p['max_depth'],
                                       n_jobs=clf_p['n_jobs'])
        elif clf_p['type'] == 'ada':
            clf = AdaBoostClassifier(n_estimators=clf_p['n_estimators'],
                                     learning_rate=clf_p['learning_rate'])
        else:
            clf = GradientBoostingClassifier(n_estimators=clf_p['n_estimators'],
                                             max_depth=clf_p['max_depth'],
                                             learning_rate=clf_p['learning_rate'],
                                             subsample=clf_p['subsample'])
    elif clf_p['clf'] == 'svm':
        mode = 'svm'
        clf = SVC(C=clf_p['C'], gamma=clf_p['gamma'], kernel=clf_p['kernel'],
                  probability=True)
    elif clf_p['clf'] == 'sgd':
        mode = 'sgd'
        clf = SGDClassifier(loss=clf_p['loss'], penalty=clf_p['penalty'],
                            alpha=clf_p['alpha'], n_iter=clf_p['n_iter'])
    else:
        raise Exception('Classifier not supported')
    return clf, mode


def compute_supervoxel_descriptor(supervoxels, descriptors, desc_type, desc_bins):
    'Mean' 'Quantized' 'Textons' 'Covar' 'Sigma Set'

    if desc_type == 'Mean':
        return spmeans(descriptors, supervoxels)
    elif desc_type == 'Covar':
        return spstats(descriptors, supervoxels, mode='add', norm=None)
    elif desc_type == 'Sigma Set':
        return spstats(descriptors, supervoxels, mode='add', sigmaset=True, norm=None)

    if desc_type == 'Textons':
        log.info('+ Applying PCA')
        descriptors = IncrementalPCA(batch_size=100).fit_transform(descriptors)

    log.info('+ Quantizing descriptors')
    cluster = MiniBatchKMeans(n_clusters=desc_bins, batch_size=100).fit_predict(descriptors)
    return sphist(cluster.astype(np.int32), supervoxels, nbins=desc_bins)


def extract_descriptors(supervoxels=None, features=None,
                        projection=None, desc_type=None, desc_bins=None,
                        nh_order=None, sp_edges=None):
    total = len(features)

    log.info('+ Reserving memory for {} features'.format(total))
    descriptors = np.zeros(DM.region_shape() + (total,), np.float32)

    for i in range(total):
        log.info('    * Loading feature {}'.format(features[i]))
        descriptors[..., i] = DM.load_slices(features[i])

    sp = None
    mask = None

    if supervoxels is not None:
        log.info('+ Loading supervoxels')
        sp = DM.load_slices(supervoxels)
        if sp.min() < 0:
            raise Exception('Supervoxels need to be recomputed for this ROI')
        descriptors.shape = (-1, total)

        num_sv = DM.attr(supervoxels, 'num_supervoxels')
        mask = np.zeros(num_sv, np.bool)
        mask[sp.ravel()] = True

        log.info('+ Computing descriptors: {} ({})'.format(desc_type, desc_bins))
        descriptors = compute_supervoxel_descriptor(sp, descriptors,
                                                    desc_type, desc_bins)
        nh_order = int(nh_order)
        if nh_order > 0:
            log.info('+ Loading edges into memory')
            edges = DM.load_ds(sp_edges)

            log.info('+ Filtering edges for selected ROI')
            idx = mask[edges[:, 0]] & mask[edges[:, 1]]
            edges = edges[idx]

            log.info('+ Aggregating neighbour features')
            G = nx.Graph()
            G.add_edges_from(edges)
            descriptors = aggregate_neighbors(descriptors, G, mode='append',
                                              norm='mean', order=nh_order)
        descriptors = descriptors[mask]

    return descriptors, sp, mask


def predict_proba(y_data=None, p_data=None,
                  level_params=None, desc_params=None,
                  clf_params=None, ref_params=None,
                  out_labels=None, out_confidence=None):

    log.info("+ Extracting descriptors: {}".format(desc_params))
    X, supervoxels, svmask = extract_descriptors(**desc_params)
    full_svmask = svmask.copy()

    log.info("+ Creating classifier: {}".format(clf_params))
    clf, mode = obtain_classifier(clf_params)

    log.info("+ Loading labels")
    labels = DM.load_slices(y_data)
    if level_params['plevel'] is not None:
        parent_labels = DM.load_slices(p_data)
        mask = parent_labels == level_params['plabel']
    else:
        mask = None

    if supervoxels is not None:
        log.info('+ Extracting supervoxel labels')
        nsp = DM.attr(desc_params['supervoxels'], 'num_supervoxels')
        nlbl = DM.attr(y_data, 'label').max()+1
        labels = _sp_labels(supervoxels.ravel(), labels.ravel(), nsp, nlbl, 0)
        if mask is not None:
            mask = mask.astype(np.int16)
            mask = np.bincount(supervoxels.ravel(), weights=mask.ravel()*2-1) > 0
        if X.shape[0] < labels.shape[0]: # less supervoxels
            labels = labels[svmask]
            if mask is not None:
                mask = mask[svmask]

    if clf is not None:
        result = _predict_proba(clf, X, labels, mask=mask, mode=mode,
                                projection=desc_params['projection'])
        X, y, pred, probs, conf, labels = result
    else:
        return None

    if supervoxels is not None:
        if ref_params['ref_type'] != 'None':
            log.info('+ Remaping supervoxels')
            svmap = np.full(nsp, -1, supervoxels.dtype)
            svmap[full_svmask] = np.arange(svmask.sum())

            log.info('+ Extracting graph')
            refine_lamda = ref_params['lambda']
            edges = DM.load_ds(ref_params['sp_edges'])
            edge_weights = DM.load_ds(ref_params['sp_eweights'])
            mean_weights = edge_weights.mean()
            edge_weights = np.minimum(edge_weights, mean_weights) / mean_weights

            log.info('+ Remapping edges')
            idx = full_svmask[edges[:, 0]] & full_svmask[edges[:, 1]]
            edges = edges[idx]
            edges[:, 0] = svmap[edges[:, 0]]
            edges[:, 1] = svmap[edges[:, 1]]
            edge_weights = edge_weights[idx]

            log.info('  * Unary potentials')
            unary = (-np.ma.log(probs)).filled()
            mapping = np.zeros(y.max()+1, np.int32)
            mapping[labels] = np.arange(labels.size)
            idx = np.where(y > -1)[0]
            col = mapping[y[idx]]
            unary[idx, col] = 0
            pairwise = edge_weights.astype(np.float32)

            log.info('  * Pairwise potentials')
            if ref_params['ref_type'] == 'Appearance':
                log.info('+ Extracting descriptive weights')
                dists = np.sqrt(np.sum((X[edges[:, 0]] - X[edges[:, 1]])**2, axis=1))
                pairwise *= np.exp(-dists.mean() * dists)

            log.debug("  * Shapes: {}, {}, {}, {}".format(unary.shape, pairwise.shape,
                                                          edges.shape, edge_weights.shape))
            log.debug("  * Edges: min: {}, max: {}".format(edges.min(0), edges.max(0)))
            log.debug("  * Unary: min: {}, max: {}, mean: {}"
                      .format(unary.min(0), unary.max(0), unary.mean(0)))
            log.debug("  * Pairwise: min: {}, max: {}, mean: {}"
                      .format(pairwise.min(), pairwise.max(), pairwise.mean()))

            log.info('+ Refining labels')
            label_cost = np.ones((labels.size, labels.size), np.float32)
            label_cost[np.diag_indices_from(label_cost)] = 0
            label_cost *= refine_lamda

            if label_cost.shape[0] == 2:
                refined = solve_binary(edges, unary, pairwise, label_cost)
            else:
                refined = solve_aexpansion(edges, unary, pairwise, label_cost)
            pred[mask] = labels[refined[mask]]

            if mask is not None:
                pred[~mask] = -1
                conf[~mask] = -1
        else:
            pass #TODO pixel refinement

        log.info('+ Mapping predictions back to pixels')
        pred_map = np.empty(nsp, dtype=pred.dtype)
        conf_map = np.empty(nsp, dtype=conf.dtype)
        pred_map[full_svmask] = pred
        conf_map[full_svmask] = conf
        pred = pred_map[supervoxels]
        conf = conf_map[supervoxels]

    pred.shape = DM.region_shape()
    conf.shape = DM.region_shape()

    log.info('+ Saving results to disk')
    DM.create_empty_dataset(out_labels, shape=DM.data_shape, dtype=pred.dtype)
    DM.write_slices(out_labels, pred, params=dict(labels=labels, active=True))

    DM.create_empty_dataset(out_confidence, shape=DM.data_shape, dtype=conf.dtype)
    DM.write_slices(out_confidence, conf, params=dict(labels=labels, active=True))

    return out_labels, out_confidence, labels


def _predict_proba(clf, X, y, mask=None, projection=None, mode=None):
    y.shape = -1

    spmask = None if mask is None else mask.ravel()
    labels = np.asarray(list(set(np.unique(y[spmask])) - set([-1])), np.int32)

    if mask is not None:
        mask.shape = -1
        idx_train = (y > -1) & mask
        idx_test = (y == -1) & mask
    else:
        idx_train = (y > -1)
        idx_test = (y == -1)

    if projection != None and projection != 'None':
        log.info('+ Projecting features')
        X = project_features(X[idx_train], projection).transform(X).astype(np.float32)

    log.info('+ Training classifier')
    clf.fit(X[idx_train], y[idx_train])

    log.info('+ Predicting labels')
    y_pred = np.full(y.size, -1, dtype=y.dtype)
    y_pred[idx_test] = clf.predict(X[idx_test])
    y_pred[idx_train] = y[idx_train]

    log.info('+ Extracting probabilities')
    y_probas = np.full((y.size, labels.size), -1, dtype=X.dtype)
    y_uncertain = np.full(y.size, -1, dtype=X.dtype)
    probas = clf.predict_proba(X[idx_test])
    y_probas[idx_test] = probas

    log.info('+ Measuring uncertainty')
    uncertain = entropy(probas.T)
    uncertain -= uncertain.min()
    uncertain /= uncertain.max()

    y_uncertain[idx_train] = 1
    y_uncertain[idx_test] = 1 - uncertain

    return X, y, y_pred, y_probas, y_uncertain, labels
